#!/usr/bin/ruby
# coding:utf-8
require "pp"
require "date"
require "time"
require "fileutils"
require "yaml"
require "logger"
require "csv"
require "securerandom"
require "optparse"

$:.unshift File.dirname(__FILE__)

require "jma/certtool"
require "jma/certtool/ca/resource"
require "jma/certtool/ca/issuingauthority"
require "jma/certtool/ca/profile"

class OrcaCA

  INDEX_KEYS = [
    :serial,
    :subject,
    :not_before,
    :not_after,
    :profile,
    :revoked,
    :revoked_at,
    :description,
  ]

  def initialize(conf)
    @conf = conf
    @ca_conf = @conf[:ca_conf]
    @ca_conf[:cert_validity] = (Date.new(2037,12,31) - DateTime.now.to_date).to_i
    @ca_conf[:subject].each do |a|
      @org = a[1] if a[0] == "O"
    end
    @org ||= "orca"
    @logger = conf[:logger]
    if JMA::CertTool::CA::IssuingAuthority.list.index(@ca_conf[:name])
      @ca = JMA::CertTool::CA::IssuingAuthority.open(@ca_conf[:name])
    else
      @logger.info("== initialize ca")
      @ca = JMA::CertTool::CA::IssuingAuthority.create(@ca_conf)
      @ca.create_profile(JMA::CertTool::CA::Profile::SSL_SERVER_PROFILE)
      @ca.create_profile(JMA::CertTool::CA::Profile::SSL_CLIENT_PROFILE)
    end
    read_index
  end

  def issue_server_cert(argv)
    @logger.info("== issue_server_cert")
    o = {:outdir => 'certs'}
    opt = OptionParser.new
    opt.on('-c','--cn=VAL','common name')          {|v| o[:cn] = v}
    opt.on('-i','--ipaddr=VAL','ip address')       {|v| o[:ip] = v}
    opt.on('-o','--outdir=VAL','output dir')       {|v| o[:outdir] = v }
    opt.parse!(argv)
    unless o[:cn] || o[:ip]
      puts "specify cn"
      puts
      puts opt.help
      exit 1
    end

    cn  = o[:cn] || o[:ip]
    dir = o[:outdir]

    FileUtils.mkdir_p(dir)
    subject = [["C","JP"],["O",@org],["CN",cn]]

    info = {
      :profile     => "ssl-server",
      :revoked     => "false",
      :revoked_at  => "",
      :description => "",
    }
    domain = o[:cn]
    ip     = o[:ip]
    args = {
     :key_algorithm    => "RSA", 
     :key_length       => 4096,
     :digest_algorithm => "SHA256",
     :subject          => subject,
     :domain           => domain,
     :ip               => ip
    }
    csr = @ca.issue_csr(args)
    cert = @ca.issue_cert({
      :profile  => info[:profile], 
      :password => @ca_conf[:password],
      :validity => @conf[:server_cert_validity], 
      :key_id   => csr.key_id
    })

    # encode! for YAML.dump
    info[:serial]     = cert.serial.to_s.encode!('us-ascii')
    info[:subject]    = cert.subject.to_s.encode!('us-ascii')
    info[:not_before] = cert.not_before.to_s.encode!('us-ascii')
    info[:not_after]  = cert.not_after.to_s.encode!('us-ascii')

    @logger.info("issue_cert: #{info}")

    _export_cert([info],dir,{:passphrase => ""})
    @index << info
    write_index
  end

  def issue_client_cert(argv)
    @logger.info("== issue_client_cert")
    o = {:outdir => 'certs',:num => 1}
    opt = OptionParser.new
    opt.on('-c','--cn=VAL','common name')          {|v| o[:cn] = v}
    opt.on('-d','--description=VAL','description') {|v| o[:desc] = v }
    opt.on('-o','--outdir=VAL','output dir')       {|v| o[:outdir] = v }
    opt.parse!(argv)
    unless o[:cn]
      puts "specify cn"
      puts
      puts opt.help
      exit 1
    end

    cn  = o[:cn]
    dir = o[:outdir]

    FileUtils.mkdir_p(dir)
    subject = [["C","JP"],["O",@org],["CN",cn]]

    info = {
      :profile     => "ssl-client",
      :revoked     => "false",
      :revoked_at  => "",
      :description => "",
    }
    args = {
     :key_algorithm    => "RSA", 
     :key_length       => 4096,
     :digest_algorithm => "SHA256",
     :subject          => subject,
    }
    csr = @ca.issue_csr(args)
    cert = @ca.issue_cert({
      :profile  => info[:profile], 
      :password => @ca_conf[:password],
      :validity => @conf[:client_cert_validity], 
      :key_id   => csr.key_id
    })

    # encode! for YAML.dump
    info[:serial]     = cert.serial.to_s.encode!('us-ascii')
    info[:subject]    = cert.subject.to_s.encode!('us-ascii')
    info[:not_before] = cert.not_before.to_s.encode!('us-ascii')
    info[:not_after]  = cert.not_after.to_s.encode!('us-ascii')

    @logger.info("issue_cert: #{info}")

    _export_cert([info],dir)
    @index << info
    write_index
  end

  def export_cert(argv)
    @logger.info("== export_cert")
    o = {:outdir => 'certs'}
    opt = OptionParser.new
    opt.on('-s','--serial=VAL','serial no string 1,2,3-7') {|v| o[:serial] = v}
    opt.on('-o','--outdir=VAL','output dir')               {|v| o[:outdir] = v }
    opt.parse!(argv)

    unless o[:serial]
      puts "specify serial"
      puts
      puts opt.help
      exit 1
    end

    list = parse_serial_str(o[:serial])
    if list.empty?
      msg = "invalid serial no string:#{o[:serial]}"
      puts msg
      @logger.info msg
      return
    end

    info_list = []
    @index.each do |i|
      if list.include?(i[:serial])
        info_list << i
        @logger.info("export #{i}")
      end
    end
    _export_cert(info_list,o[:outdir])
  end

  def find_and_show_cert_for_revoke(serial)
    info = nil
    @index.each do |i|
      if i[:serial] == serial
        info = i
        INDEX_KEYS.each do |k|
          puts "#{k}: #{i[k]}"
        end
        break
      end
    end
    return info 
  end

  def revoke_cert(argv)
    @logger.info("== revoke_cert")

    serial = nil
    opt = OptionParser.new
    opt.on('-s','--serial=VAL','serial no') {|v| serial = v}
    opt.parse!(argv)

    unless serial
      puts "specify serial"
      puts
      puts opt.help
      exit 1
    end

    info = find_and_show_cert_for_revoke(serial)
    unless info
      msg = "no cert,serial:#{serial}"
      puts msg
      @logger.info(msg)
      return
    end
    puts
    print "really revoke?(y or n):"
    if /^[y|Y]/ =~ STDIN.gets
      @ca.revoke_cert(serial.to_i)
      puts "revoked"
      info[:revoked] = "true"
      info[:revoked_at] = Time.now.to_s
      write_index
      @logger.info("revoked #{info}")
    else
      puts "canceled"
      @logger.info("canceled")
    end
  end

  def issue_crl(argv)
    @logger.info("== issue_crl")
    filename = nil
    opt = OptionParser.new
    opt.on('-f','--filename=VAL','output filename') {|v| filename = v}
    opt.parse!(argv)

    unless filename
      puts "specify filename"
      puts
      puts opt.help
      exit 1
    end

    crl = @ca.issue_crl({
      :password => @ca_conf[:password],
      :validity => @ca_conf[:crl_validity]})
    File.open(filename,"w") do |io|
      io.print crl.to_pem
    end
    msg = "issued and exported to #{filename}"
    puts msg
    @logger.info(msg)
  end

  def list_cert_usage
    puts "== list_cert usage"
    puts "#{File.basename($0)} list_cert"
    puts "#{File.basename($0)} list_cert enable"
    puts "#{File.basename($0)} list_cert expired"
    puts "#{File.basename($0)} list_cert revoked"
    puts "#{File.basename($0)} list_cert will_expire <Time:2017-01-01 00:00:00 +9:00 or now> <Time:2017-01-01 00:00:00 +9:00>"
    puts "#{File.basename($0)} list_cert subject <\"/C=JP/O=org/CN=name\" or \"O=org\">"
    puts "#{File.basename($0)} list_cert serial <1,2,3-10>"
    puts "#{File.basename($0)} list_cert description <desc>"
  end

  def display_cert(c)
    a = [
     c[:serial],
     c[:subject],
     c[:not_before],
     c[:not_after],
     c[:revoked],
     c[:revoked_at],
     c[:profile],
     c[:description],
    ]
    puts a.join(",")
  end

  def list_cert(args)
    @logger.info("== list_cert #{args}")
    if args.empty?
      @index.each do |c|
        display_cert(c)
      end
      return
    end
    case args[0]
    when "enable"
      t = Time.now
      list = @index.select do |c|
        t1 = Time.parse(c[:not_before])
        t2 = Time.parse(c[:not_after])
        t1 < t && t < t2 && c[:revoked] != "true"
      end
      list.each do |c|
        display_cert(c)
      end
    when "expired"
      t = Time.now
      list = @index.select do |c|
        t2 = Time.parse(c[:not_after])
        t2 < t && c[:revoked] != "true"
      end
      list.each do |c|
        display_cert(c)
      end
    when "revoked"
      list = @index.select do |c|
        c[:revoked] == "true"
      end
      list.each do |c|
        display_cert(c)
      end
    when "will_expire"
      list_cert_usage if args.size < 3
      if args[1] == "now"
        t1 = Time.now
      else
        t1 = Time.parse(args[1])
      end
      if args[2] == "now"
        t2 = Time.now
      else
        t2 = Time.parse(args[2])
      end
      t1, t2 = t2, t1 if t1 > t2

      list = @index.select do |c|
        t = Time.parse(c[:not_after])
        t1 < t && t < t2 && c[:revoked] != "true"
      end
      list.each do |c|
        display_cert(c)
      end
    when "subject"
      list_cert_usage unless args[1]
      list = @index.select do |c|
        c[:subject].include?(args[1])
      end
      list.each do |c|
        display_cert(c)
      end
    when "serial"
      list_cert_usage unless args[1]
      s_list = parse_serial_str(args[1])
      list = @index.select do |c|
        s_list.include?(c[:serial])
      end
      list.each do |c|
        display_cert(c)
      end
    when "description"
      list_cert_usage unless args[1]
      list = @index.select do |c|
        c[:description].include?(args[1])
      end
      list.each do |c|
        display_cert(c)
      end
    else
      list_cert_usage
      return
    end
  end

  private

  def _export_cert(list,dir,opts = {})
    FileUtils.mkdir_p(dir)
    list.each do |info|
      make_cert_archive(dir,info,opts)
    end
    msg = "exported to #{File.expand_path(dir)}"
    puts msg
    @logger.info(msg)
  end

  def make_cert_archive(dir,info,opts = {})
    serial  = info[:serial].to_i
    subject = info[:subject]
    _subject = subject.gsub('/','').gsub(%r!C=|O=|CN=!,'_')
    base = "%06d_#{_subject}"%serial
    abdir = File.expand_path(dir)
    tmpdir do |tmp|
      zipdir = File.join(tmp,base)
      FileUtils.mkdir_p(zipdir)

      certfile = File.join(zipdir,"#{base}.crt")
      File.open(certfile,"w"){|io| io.print @ca.export_cert(serial)[:certificate].to_pem}

      keyfile = File.join(zipdir,"#{base}.pem")
      File.open(keyfile,"w"){|io| io.print @ca.read_cert_key(serial,"").to_pem}

      size = 16
      pass = opts[:passphrase]
      unless pass
        pass = ""
        chars = ('a'..'z').to_a +
           ('A'..'Z').to_a + ('0'..'9').to_a - ['l','1','I','0','O']
        size.times do
          pass << chars[rand(chars.length)]
        end
      end

      pkcs12 = File.join(zipdir,"#{base}.p12")
      args = {
          :serial          => serial,
          :load_password   => "",
          :export_password => pass
      }
      der = @ca.export_pkcs12(args).to_der
      File.open(pkcs12,"wb"){|io| io.print der }

      passfile = File.join(zipdir,"#{base}.pass")
      File.open(passfile,"w"){|io| io.puts pass}

      key = File.join(zipdir,"#{base}.enc.pem")
      if pass.empty?
        File.open(key,"w"){|io| io.print @ca.read_cert_key(serial,"").to_pem}
      else
        File.open(key,"w"){|io| io.print @ca.read_cert_key(serial,"").to_pem(OpenSSL::Cipher.new("AES-256-CBC"),pass)}
      end

      cacert = File.join(zipdir,"ca.crt")
      File.open(cacert,"w"){|io| io.print @ca.export_ca_cert.to_pem}

      zipfile = "#{base}.zip"
      `zip -r #{zipfile} #{base}`
      FileUtils.cp(zipfile,abdir)
    end
  end

  def parse_serial_str(serial_str)
    list = []
    serial_str.split(",").each do |item|
      if /^(\d+)$/ =~ item
        list << $1
      elsif /^(\d+)-(\d+)$/ =~ item
        r1 = $1.to_i
        r2 = $2.to_i
        r1, r2 = r2, r1 if r1 > r2
        list << (r1..r2).to_a.map do |e| e.to_s end
        list.flatten!
      end
    end
    list
  end

  def make_tmpdir
    begin
      dir = File.join('/tmp',SecureRandom.uuid)
      FileUtils.mkdir(dir)
      return dir
    rescue Errno::EEXIST
      retry
    end
  end

  def tmpdir
    begin
      dir = make_tmpdir
      orgdir = FileUtils.pwd
      FileUtils.chdir(dir)
      yield(dir)
      FileUtils.chdir(orgdir)
    ensure
      FileUtils.rm_rf(dir) if dir
    end
  end

  def read_index
    file = File.join(@conf[:storedir],"index.yml")
    @serial = 2
    if File.exist?(file)
      @index = YAML.load(File.read(file))
      @index.each do |i|
        s = i[:serial].to_i + 1
        @serial = s if @serial < s
      end
      #@index.reject! do |c|
      #  c[:profile] == "ssl-server"
      #end
    else
      @index = []
    end
  end

  def write_index
    file = File.join(@conf[:storedir],"index.yml")
    File.open(file,"w") do |io|
      io.print(YAML.dump(@index))
    end
  end

  def subject_to_a(dn)
    h = {} 
    dn.split("/").each do |elem|
      h[$1] = $2 if %r|\s*(\w+)\s*=\s*(.+)| =~ elem
    end
    return nil unless h["C"] == "JP" 
    return nil unless h["O"]
    return nil unless h["CN"]
    return h.to_a
  end

end

def lock_exclusive
  path = File.join(File.expand_path(ENV["HOME"]),".orca_catool.lock")
  lock_file = open(path,File::RDWR|File::CREAT) 
  begin
    yield if lock_file.flock(File::LOCK_EX)
  ensure
    lock_file.close
  end
end

def usage
  puts <<-EOS
  specify subcommand

  issue_server_cert
  issue_client_cert
  export_cert
  revoke_cert
  issue_crl
  list_cert
  EOS
  exit 1
end

# main
lock_exclusive do
  DEFAULT_ROOT=File.join(File.expand_path(ENV["HOME"]),".orca_catool")
  root = ENV["ORCA_CATOOL_ROOT"] || DEFAULT_ROOT
  conffile = File.join(root,"config.yml")
  unless File.exists?(conffile)
    puts "config file(#{conffile}) does not exist."
    puts
    puts "initialize ca?"
    print "y or n> "
    a = STDIN.gets
    if /^y/ =~ a
      FileUtils.mkdir_p(root)
      tmpldir = File.dirname(File.expand_path(__FILE__))
      tmpl = File.join(tmpldir,"config.yml.sample")
      FileUtils.cp(tmpl,conffile)
    else
      exit 0
    end
  end

  conf = YAML.load(File.read(conffile))
  conf[:rootdir] = root
  JMA::CertTool::CA::Resource.root = conf[:rootdir]

  logdir = File.join(conf[:rootdir],"log")
  FileUtils.mkdir_p(logdir)
  conf[:logger] = Logger.new(File.join(logdir,"ca.log"),"monthly")
  conf[:storedir] = File.join(conf[:rootdir],conf[:ca_conf][:name])
  ca = OrcaCA.new(conf)

  usage if ARGV.size < 1

  command = ARGV.shift
  case command
  when "issue_server_cert"
    ca.issue_server_cert(ARGV)
  when "issue_client_cert"
    ca.issue_client_cert(ARGV)
  when "export_cert"
    ca.export_cert(ARGV)
  when "revoke_cert"
    ca.revoke_cert(ARGV)
  when "issue_crl"
    ca.issue_crl(ARGV)
  when "list_cert"
    ca.list_cert(ARGV)
  else
    usage
  end
end
