require "jma/certtool/ca"
require "jma/certtool/ca/resource"
require "jma/certtool/ca/x509"
require "jma/certtool/openssl"
require "pp"

class JMA::CertTool::CA::IssuingAuthority
  include JMA::CertTool::CA::Utilities
  @@resource_class = JMA::CertTool::CA::Resource

  class <<self
    def create(property)
      name = property[:name] || "default"
      if list.member?(name)
        raise RuntimeError, "CA named `%s' already exists" % name
      end
      ca = new(name)
      ca.setup_resource
      ca.issue_self_signed(property)
      ca.issue_crl(property)
      return ca
    end

    def delete(name)
      resource = @@resource_class.new(name)
      resource.delete
    end

    def list
      return @@resource_class.list
    end

    def resource_class=(klass)
      @@resource_class = klass
    end

    alias open new
  end

  attr_reader :name

  def initialize(name)
    @name = name
    @resource = @@resource_class.new(name)
  end

  def delete
    @resource.delete
  end

  ### CA operatins
  def setup_resource
    @resource.create
  end

  def issue_self_signed(property)
    subject = create_dn(property[:subject])

    key_algorithm = property[:key_algorithm]
    key_length = property[:key_length]
    key = JMA::CertTool::CA::X509.generate_key(key_algorithm, key_length)

    serial = @resource.checkout_serial

    profile = OpenSSL::Config.new
    profile.validity = property[:cert_validity]
    profile.extensions = property[:cert_extensions] || []
    profile.digest_algorithm = property[:digest_algorithm]
    cert = JMA::CertTool::CA::X509.issue_self_signed(
      subject, key, serial, profile, {}
    )
    @resource.store_ca_key(key, property[:password])
    @resource.store_property(property)
    @resource.store_ca_cert(cert)

    return cert
  end

  def issue_csr_of_ca_key(arguments)
    subject = @resource.load_ca_cert.subject
    key = @resource.load_ca_key(arguments[:password])
    digest_algorithm = @resource.load_property[:digest_algorithm]
    return JMA::CertTool::CA::X509.issue_csr(subject, key, digest_algorithm, [])
  end

  def import_ca_cert(cert)
    current_cert = @resource.load_ca_cert
    if cert.key_id != current_cert.key_id
      raise RuntimeError, "Specified certificate doesn't match with CA key."
    end
    @resource.store_ca_cert(cert)
  end

  def issue_crl(arguments)
    property = @resource.load_property
    profile = OpenSSL::Config.new
    profile.validity = property[:crl_validity]
    profile.extensions = property[:crl_extensions] || []
    profile.digest_algorithm = property[:digest_algorithm]

    options = Hash.new
    options[:validity] = arguments[:validity]
    if (last_crl = @resource.latest_crl).nil?
      options[:crl_number] = 1
    elsif crl_number = last_crl.crl_number
      options[:crl_number] = crl_number + 1
    end

    issuer = @resource.load_ca_cert
    key = @resource.load_ca_key(arguments[:password])
    revoked = []
    @resource.each_revoked_certificate{|attr|
      if attr[:not_after] > current_time
        serial = attr[:serial]
        time = attr[:revoke_time]
        reason = attr[:revoke_reason]
        revoked << OpenSSL::X509::Revoked.generate(serial, time, reason)
      end
    }
    crl = JMA::CertTool::CA::X509.issue_crl(issuer, key, profile, revoked, options)
    @resource.add_crl(crl)

    return crl
  end

  def export_ca_cert
    return @resource.load_ca_cert
  end

  def export_crl
    return @resource.latest_crl
  end

  def change_password(current_password, new_password)
    return @resource.change_ca_password(current_password, new_password)
  end

  ### Security Device Option
  def change_device_option(option)
    property = @resource.load_property
    option.each{ |k,v|
      property[k] = v
    }
    @resource.store_property(property)
  end

  def load_device_option
    property = @resource.load_property
    keys = [:pkcs15_init, :pkcs15_tool, :opensc_tool, 
            :puk, :sopin, :sopuk]
    property.reject!{|k,v|
      true unless keys.include?(k)
    }
    return property
  end

  ### Profile
  def create_profile(profile)
    @resource.create_profile(profile)
  end

  def read_profile(name)
    return @resource.read_profile(name)
  end

  def delete_profile(name)
    return @resource.delete_profile(name)
  end

  ### EE Certificate
  def issue_cert(arguments)
    unless csr_info = @resource.read_csr(arguments[:key_id])
      return false
    end

    csr = csr_info[:csr]
    profile = @resource.read_profile(arguments[:profile])
    options = {
      :not_before => arguments[:not_before],
      :not_after => arguments[:no_after],
      :validity => arguments[:validity],
    }
    cert = JMA::CertTool::CA::X509.issue_cert(
      @resource.load_ca_cert, @resource.load_ca_key(arguments[:password]),
      csr, @resource.checkout_serial, profile, options
    )

    @resource.create_cert(cert, 
      :user_name => arguments[:user_name],
      :profile => arguments[:profile]
    )
    @resource.update_csr(csr.key_id, :signed => true)

    return cert
  end

  def delete_cert(serial)
    cert_info = @resource.read_cert(serial)
    delete_key(cert_info[:key_id])
    @resource.delete_cert(serial)
  end

  def revoke_cert(serial, reason=OpenSSL::X509::Revoked::UNSPECIFIED)
    @resource.update_cert(serial,
      :revoke_reason => reason,
      :revoke_time => current_time,
      :revoked => true
    )
  end

  def update_cert(serial, attributes)
    @resource.update_cert(serial, attributes)
  end

  def export_cert(serial)
    return @resource.read_cert(serial)
  end

  def export_pkcs12(arguments)
    serial = arguments[:serial]
    load_pass = arguments[:load_password]
    export_pass = arguments[:export_password]

    issuer = @resource.load_ca_cert
    cert_info = @resource.read_cert(serial)
    cert = cert_info[:certificate]
    name = cert_info[:user_name]
    key = @resource.read_key(cert.key_id, load_pass)
    return OpenSSL::PKCS12.create(export_pass, name, key, cert, [issuer])
  end

  ### CSR
  def issue_csr(arguments)
    subject = create_dn(arguments[:subject])

    key_algorithm = arguments[:key_algorithm]
    key_length = arguments[:key_length]
    key = JMA::CertTool::CA::X509.generate_key(key_algorithm, key_length)
    digest_algorithm = arguments[:digest_algorithm]

    ext= []
    subject_alt_name_value = []
    if email = arguments[:email]
      email.split(",").each{|addr|
        subject_alt_name_value << "email:#{addr.strip}"
      }
    end
    if domain = arguments[:domain]
      domain.split(",").each{|name|
        subject_alt_name_value << "DNS:#{name.strip}"
      }
    end
    if ip = arguments[:ip]
      ip.split(",").each{|name|
        subject_alt_name_value << "IP:#{name.strip}"
      }
    end
    if subject_alt_name_value.length > 0
      subject_alt_name = subject_alt_name_value.join(", ")
      ef = OpenSSL::X509::ExtensionFactory.new
      ext << ef.create_extension("subjectAltName", subject_alt_name)
    end

    csr = JMA::CertTool::CA::X509.issue_csr(subject, key, digest_algorithm, ext)
    @resource.create_key(key, arguments[:password])
    @resource.create_csr(csr)

    return csr
  end

  def export_csr(key_id)
    @resource.read_csr(key_id)
  end

  def import_csr(csr)
    @resource.create_csr(csr)
  end

  def import_csr_from_file(path)
    csr = read_pem(path, OpenSSL::X509::Request)
    @resource.create_csr(csr)
  end

  def delete_csr(key_id)
    csr_info = @resource.read_csr(key_id)
    unless csr_info[:signed]
      delete_key(key_id)
    end
    @resource.delete_csr(key_id)
  end

  ### Private Key
  def key_exists?(key_id)
    return @resource.key_exists?(key_id)
  end

  def delete_key(key_id)
    @resource.delete_key(key_id)
  end

  def read_key(key_id,passwd)
    @resource.read_key(key_id,passwd)
  end

  def read_cert_key(serial,passwd)
    cert_info = @resource.read_cert(serial)
    cert = cert_info[:certificate]
    @resource.read_key(cert.key_id,passwd)
  end

  ### Iterators
  def each_certificate(&block)
    @resource.each_certificate{|cert| block.call(cert) }
  end

  def each_csr(&block)
    @resource.each_csr{|csr| block.call(csr) }
  end

  def each_profile(&block)
    @resource.each_profile{|profile| block.call(profile) }
  end

  def profiles
    result = []
    each_profile{|profile| result << profile }
    return result
  end

  private

  def create_dn(ary)
    dn = OpenSSL::X509::Name.new
    ary.each{|oid, value|
      if value && value.length > 0
        dn.add_entry(oid, value)
      end
    }
    return dn
  end
end
