require "jma/certtool/ca"
require "jma/certtool/ca/x509"
require "jma/certtool/openssl"
require "fileutils"
require "thread"
require "yaml/store"

# check for YAML option issue
#   http://diary.does.notwork.org/gotoyuzo/?date=20061001#p02
opts = {}
true.to_yaml(opts)
if opts[:KeepValue]
  class YAML::Store
    def dump(table)
      @table.to_yaml
    end
  end
end

class JMA::CertTool::CA::Resource
  include JMA::CertTool::CA::Utilities

  Directories = []

  attr_reader :name

  class <<self
    DEFAULT_ROOT = File.expand_path("~/.jma-receipt/ca")
    @@root = nil

    def root
      return @@root || DEFAULT_ROOT
    end

    def root=(dir)
      @@root = dir
    end

    def pathname(name, *relatives)
      Directories << name if /_dir$/ =~ name.to_s
      define_method(name){|*args|
        format = top_dir
        relatives.each{|rel| format = File.join(format, rel) }
        sprintf(format, *args)
      }
    end

    def list
      begin
        entries = Dir.entries(root)
        entries.reject!{|entry| /^\./ =~ entry }
        return entries.sort
      rescue Errno::ENOENT
        return []
      end
    end
  end

  #####

  pathname :property_file, "property.yml"
  pathname :ca_cert_file, "cacert.crt"
  pathname :ca_key_file, "cakey.pem"
  pathname :latest_crl_file, "latest.crl"
  pathname :serial_file, "serial"

  pathname :cert_store_file, "cert.yml"
  pathname :csr_store_file, "csr.yml"
  pathname :profile_store_file, "profile.yml"

  pathname :cert_dir, "cert"
  pathname :csr_dir, "csr"
  pathname :key_dir, "key"
  pathname :crl_dir, "crl"
  pathname :profile_dir, "profile"

  pathname :cert_file, "cert", "%010d.crt"
  pathname :csr_file, "csr", "%s.csr"
  pathname :key_file, "key", "%s.pem"
  pathname :crl_file, "crl", "%010d.crl"
  pathname :profile_file, "profile", "%s.cnf"

  ####

  def initialize(name)
    @name = name
    @root = self.class.root
    @cert_store = nil
    @csr_store = nil
  end

  def cert_store
    @cert_store ||= YAML::Store.new(cert_store_file)
  end

  def csr_store
    @csr_store ||= YAML::Store.new(csr_store_file)
  end

  def top_dir
    return File.join(@root, @name)
  end

  def create
    Directories.each{|name| FileUtils.mkdir_p(__send__(name)) }
    File.chmod(0700, @root)
  end

  def delete
    FileUtils.rm_rf(top_dir)
  end

  SerialMutex = Mutex.new

  def checkout_serial
    flag = File::RDWR|File::CREAT|File::EXCL
    begin
      SerialMutex.synchronize do
        File.open(serial_file, flag){|io|
          io.flock(File::LOCK_EX)
          serial = Integer(io.gets || 1)
          io.rewind
          io.truncate(0)
          io.puts(serial + 1)
          return serial
        }
      end
    rescue Errno::EEXIST
      flag &= ~File::EXCL
      retry
    end
  end

  ####

  def store_property(prop)
    prop = prop.dup
    prop.delete(:password)
    write_file(property_file, YAML.dump(prop))
  end

  def load_property
    return YAML.load(File.read(property_file))
  end

  def store_ca_cert(cert)
    write_pem(ca_cert_file, cert)
  end

  def load_ca_cert
    read_pem(ca_cert_file, OpenSSL::X509::Certificate)
  end

  def store_ca_key(key, password)
    return JMA::CertTool::CA::X509.store_key_to_file(ca_key_file, key, password)
  end

  def load_ca_key(password)
    return JMA::CertTool::CA::X509.load_key_from_file(ca_key_file, password)
  end

  def change_ca_password(current_password, new_password)
    key = JMA::CertTool::CA::X509.load_key_from_file(ca_key_file, current_password)
    JMA::CertTool::CA::X509.store_key_to_file(ca_key_file, key, new_password)
  end

  def add_crl(crl)
    crl_number = crl.crl_number || current_time.to_i
    [crl_file(crl_number), latest_crl_file].each{|file|
      write_pem(file, crl)
    }
  end

  def latest_crl
    begin
      return read_pem(latest_crl_file, OpenSSL::X509::CRL)
    rescue
      return nil
    end
  end

  def load_crl(crl_number)
    return read_pem(crl_file(crl_number), OpenSSL::X509::CRL)
  end

  def delete_crl(crl_number)
    FileUtils.rm_f(crl_file(crl_number))
  end

  ####

  def set_default_cert_attribute(cert, attr)
    subject = cert.subject
    common_name = subject.to_a.assoc("CN")[1]
    return attr.dup.update({
      :serial => cert.serial.to_i,
      :subject => subject.to_s(OpenSSL::X509::Name::RFC2253),
      :common_name => common_name,
      :not_before => cert.not_before,
      :not_after => cert.not_after,
      :key_id => cert.key_id,
    })
  end

  def create_cert(cert, attribute={})
    attribute = set_default_cert_attribute(cert, attribute)
    cert_store.transaction{
      write_pem(cert_file(cert.serial), cert)
      cert_store[cert.serial.to_i] = attribute
    }
    return true
  end

  def read_cert(serial)
    cert_store.transaction{
      if attr = cert_store[serial]
        attr = attr.dup
        attr[:certificate] =
          read_pem(cert_file(serial), OpenSSL::X509::Certificate)
        return attr
      end
      return nil
    }
  end

  def update_cert(serial, attribute={})
    cert_store.transaction{
      if attr = cert_store[serial]
        attr.update(attribute)
        cert_store[serial] = attr
      end
    }
  end

  def delete_cert(serial)
    cert_store.transaction{
      FileUtils.rm_f(cert_file(serial))
      cert_store.delete(serial)
    }
  end

  ####

  def set_default_csr_attribute(csr, attribute)
    key = csr.public_key
    key_algorithm, key_length = case key
    when OpenSSL::PKey::RSA
      ["RSA", key.n.num_bits]
    when OpenSSL::PKey::DSA
      ["DSA", key.p.num_bits]
    else
      ["Unknown", 0]
    end
    subject = csr.subject
    common_name = subject.to_a.assoc("CN")[1]
    ret = attribute.dup.update({
      :subject => subject.to_s(OpenSSL::X509::Name::RFC2253),
      :common_name => common_name,
      :key_algorithm => key_algorithm,
      :key_length => key_length,
      :request_date => current_time,
      :key_id => csr.key_id,
    })
    unless ret.has_key?(:signed)
      ret[:signed] = false
    end
    return ret
  end

  def create_csr(csr, attribute={})
    attribute = set_default_csr_attribute(csr, attribute)
    csr_store.transaction{
      write_pem(csr_file(csr.key_id), csr)
      csr_store[csr.key_id] = attribute
    }
    return true
  end

  def read_csr(key_id)
    csr_store.transaction{
      if attr = csr_store[key_id]
        attr = attr.dup
        attr[:csr] = read_pem(csr_file(key_id), OpenSSL::X509::Request)
      end
      return attr
    }
  end

  def update_csr(key_id, attribute={})
    csr_store.transaction{
      if attr = csr_store[key_id]
        attr.update(attribute)
        csr_store[key_id] = attr
      end
    }
  end

  def delete_csr(key_id)
    csr_store.transaction{
      FileUtils.rm_f(csr_file(key_id))
      csr_store.delete(key_id)
    }
  end

  ####

  def create_key(key, password)
    JMA::CertTool::CA::X509.store_key_to_file(key_file(key.key_id), key, password)
  end

  def read_key(key_id, password)
    return JMA::CertTool::CA::X509.load_key_from_file(key_file(key_id), password)
  end

  def key_exists?(key_id)
    return File.file?(key_file(key_id))
  end

  def delete_key(key_id)
    FileUtils.rm_f(key_file(key_id))
  end

  ####

  def create_profile(profile)
    write_file(profile_file(profile.name), profile.dump)
  end

  def read_profile(name)
    return OpenSSL::Config.parse(File.read(profile_file(name)))
  end

  def update_profile(name, &block)
    profile = read_profile(name)
    block.call(profile)
    create_profile(profile)
  end

  def delete_profile(name)
    FileUtils.rm_f(profile_file(name))
  end

  ####

  def each_certificate(&block)
    cert_store.transaction{
      roots = cert_store.roots
      roots.sort.each{|serial| block.call(cert_store[serial]) }
    }
  end

  def each_revoked_certificate(&block)
    now = current_time
    each_certificate{|cert|
      block.call(cert) if cert[:revoked]
    }
  end

  def each_csr(&block)
    csr_store.transaction{
      roots = csr_store.roots
      roots.sort.each{|key_id| block.call(csr_store[key_id]) }
    }
  end

  def each_profile(&block)
    entries = Dir.entries(profile_dir)
    entries.reject!{|entry| /\.cnf$/ !~ entry }
    entries.collect!{|entry| entry.sub(/\.cnf$/, "") }
    entries.sort.each{|entry| block.call(read_profile(entry)) }
  end
end
