require "jma/certtool/ca"
require "jma/certtool/openssl"

module JMA::CertTool::CA::X509
  X509Error = Class.new(StandardError)
  UnsupportedKeyAlgorithm = Class.new(X509Error)
  UnknownRevokeReason = Class.new(X509Error)
  InvalidCertValidity = Class.new(X509Error)
  InvalidCRLValidity = Class.new(X509Error)

  module_function

  def load_key_from_file(file, password)
    pem = File.read(file)
    begin
      return OpenSSL::PKey::RSA.new(pem, password)
    rescue
      return OpenSSL::PKey::DSA.new(pem, password)
    end
  end

  def store_key_to_file(file, key, password)
    File.open(file, "w"){|io|
      if password && password.length > 0
        cipher = nil
        cipher ||= OpenSSL::Cipher::AES.new(256, "CBC")
        cipher ||= OpenSSL::Cipher::DES.new
        io.write(key.export(cipher, password))
      else
        io.write(key.to_pem)
      end
    }
  end

  def issue_csr(subject_dn, key, digest_algorithm, extensions)
    csr = OpenSSL::X509::Request.new
    csr.version = 0   # CSRv1
    csr.subject = subject_dn
    csr.public_key = key.public_key
    if extensions && extensions.size > 0
      # Add extensions as extension-request attribute.
      # See also: PKCS #9 or RFC2985
      ext = OpenSSL::ASN1::Set([OpenSSL::ASN1::Sequence(extensions)])
      attr = OpenSSL::X509::Attribute.new("extReq", ext)
      csr.add_attribute(attr)
    end
    do_sign(csr, key, digest_algorithm)

    return csr
  end

  def check_cert_validity(cert, issuer_cert, max_days)
    if cert.not_before >= cert.not_after
      raise InvalidCertValidity, "not_before >= not_after"
    end
    if cert.not_after - cert.not_before > max_days * 86400
      raise InvalidCertValidity, "validity too long"
    end
    if cert.not_before < issuer_cert.not_before
      # "RFC2510 -- 2.4 Root CA Key update" says:
      #
      #  The "old with new" certificate must have a validity period starting
      #  at the generation time of the old key pair and ending at the expiry
      #  date of the old public key.
      #
      # We should issue certificates which has start time earlier than
      # the issuer's start time.
      if cert.subject != issuer_cert.subject
        raise InvalidCertValidity, "not_before < issuer.not_before"
      end
    end
    if cert.not_after > issuer_cert.not_after
      raise InvalidCertValidity, "not.after > issuer.not_after"
    end
  end

  def issue_cert(issuer_cert, issuer_key, csr, serial, profile, options)
    now = JMA::CertTool::CA::Utilities.current_time
    cert = OpenSSL::X509::Certificate.new
    issuer_cert = cert unless issuer_cert

    cert.version = 2   # X509v3
    cert.serial = serial
    cert.public_key = csr.public_key
    cert.subject = csr.subject
    cert.issuer = issuer_cert.subject

    validity = options[:validity] || Integer(profile.validity)
    cert.not_before = options[:not_before] || now
    cert.not_after = options[:not_after] || cert.not_before + validity * 86400
    check_cert_validity(cert, issuer_cert, validity)

    ef = OpenSSL::X509::ExtensionFactory.new
    ef.issuer_certificate = issuer_cert
    ef.subject_certificate = cert
    ef.config = profile
    ext_req = csr.extensions
    profile.extensions.each{|oid, value|
      if value.nil? || value.empty?
        next
      elsif value == "copy_from_request"
        exts = ext_req.find_all{|e| e.oid == oid }
        exts.each{|ext| cert.add_extension(ext) }
      else
        ext = ef.create_extension(oid, value)
        cert.add_extension(ext)
      end
    }
    do_sign(cert, issuer_key, profile.digest_algorithm)

    return cert
  end

  def issue_self_signed(subject_dn, key, serial, profile, options)
    csr = issue_csr(subject_dn, key, profile.digest_algorithm, nil)
    return issue_cert(nil, key, csr, serial, profile, options)
  end

  def issue_link_certificates(old_with_old, old_key,
                              new_with_new, new_key,
                              serial_for_old_with_new,
                              serial_for_new_with_old,
                              profile)
    issuer  = new_with_new
    key     = new_key
    digest = profile.digest_algorithm
    csr     = issue_csr(old_with_old.subject, old_key,
                        profile.digest_algorithm, nil)
                        
    serial  = serial_for_old_with_new
    options = {
      :not_before => old_with_old.not_before,
      :not_after => old_with_old.not_after,
    }
    old_with_new = issue_cert(issuer, key, csr, serial, profile, options)

    issuer  = old_with_old
    key     = old_key
    csr     = issue_csr(new_with_new.subject, new_key,
                        profile.digest_algorithm, nil)
    serial  = serial_for_new_with_old
    options = {
      :not_before => new_with_new.not_before,
      :not_after => old_with_old.not_after,
    }
    new_with_old = issue_cert(issuer, key, csr, serial, profile, options)

    return [old_with_new, new_with_old]
  end

  def issue_crl(issuer, issuer_key, profile, revoke_info, options)
    now = JMA::CertTool::CA::Utilities.current_time
    crl = OpenSSL::X509::CRL.new
    crl.version = 1
    crl.issuer = issuer.subject
    validity = options[:validity] || Integer(profile.validity)
    crl.last_update = options[:this_update] || options[:last_update] || now
    crl.next_update = options[:next_update] || crl.last_update + validity * 86400
    if crl.last_update >= crl.next_update
      raise InvalidCRLValidity, "last_update >= next_update"
    end
    revoke_info.each{|info| crl.add_revoked(info) }
    ef = OpenSSL::X509::ExtensionFactory.new
    ef.issuer_certificate = issuer
    ef.crl = crl
    ef.config = profile
    profile.extensions.each{|oid, value|
      crl.add_extension(ef.create_extension(oid, value))
    }
    if serial = options[:crl_number] || options[:serial]
      crl.crl_number = serial
    end
    do_sign(crl, issuer_key, profile.digest_algorithm)

    return crl
  end

  def generate_key(algorithm, length)
    case algorithm
    when /\Arsa\z/i 
      return OpenSSL::PKey::RSA.new(length)
    when /\Adsa\z/i
      return OpenSSL::PKey::DSA.new(length)
    else
      message = "%s is not supprted" % algorithm.inspect
      rasie UnsupportedKeyAlgorithm, message
    end
  end

  def do_sign(obj, key, digest_algorithm)
    digest = nil
    case key
    when OpenSSL::PKey::RSA
      if %w(MD5 SHA1 SHA256).member?(digest_algorithm)
        digest = OpenSSL::Digest.new(digest_algorithm)
      end
    when OpenSSL::PKey::DSA
      if %w(SHA1 DSS1).member?(digest_algorithm)
        digest = OpenSSL::Digest::DSS1.new("DSS1")
      end
    else
      message = "%s is not supprted" % algorithm.inspect
      rasie UnsupportedKeyAlgorithm, message
    end
    obj.sign(key, digest || key.default_digest)
  end
end
