require "openssl"

module OpenSSL
  module PKey
    DEFAULT_GENERATE_CALLBACK = Proc.new{|p, n|
      case p
      when 0 then $stderr.putc "."  # BN_generate_prime
      when 1 then $stderr.putc "+"  # BN_generate_prime
      when 2 then $stderr.putc "*"  # searching good prime, n = # of try,
                                    # but also data from BN_generate_prime
      when 3 then $stderr.putc "\n" # BN_generate_prime, n = 0 - p, n = 1 - q
                                    # but also data from BN_generate_prime)
      else $stderr.putc "*"         # BN_generate_prime
      end
    }

    module PublicKeyOperation
      def key_id
        return OpenSSL::Digest::SHA1.hexdigest(public_key.to_der)
      end
    end

    class PKey
      include ::OpenSSL::PKey::PublicKeyOperation
    end

    class RSA
      def ==(other)
        return false if self.class != other.class
        return false if self.private? != other.private?
        return false if self.n != other.n
        return false if self.e != other.e
        return true
      end

      def default_digest
        return OpenSSL::Digest::SHA1.new
      end
    end

    class DSA
      def ==(other)
        return false if self.class != other.class
        return false if self.private? != other.private?
        return false if self.p != other.p
        return false if self.q != other.q
        return false if self.g != other.g
        return false if self.pub_key != other.pub_key
        return true
      end

      def default_digest
        return OpenSSL::Digest::DSS1.new
      end
    end
  end

  module X509
    class Certificate
      include OpenSSL::PKey::PublicKeyOperation

      def digest(digest=nil)
        (digest || default_digest).digest(to_der)
      end

      def hexdigest(digest=nil)
        (digest || default_digest).hexdigest(to_der)
      end
    end

    class Request
      include OpenSSL::PKey::PublicKeyOperation

      def extensions
        ext = []
        self.attributes.each{|attr|
          if attr.oid == "extReq" || attr.oid == "msExtReq"
            seq = attr.value.value[0]
            seq.value.each{|a1ext|
              ext << Extension.new(a1ext)
            }
          end
        }
        return ext
      end
    end

    class CRL
      def crl_number
        self.extensions.each{|ext|
          if ext.oid == "crlNumber"
            return Integer(ext.value)
          end
        }
        return nil
      end
      alias serial crl_number

      def crl_number=(num)
        crlnum = OpenSSL::ASN1::Integer(num)
        self.add_extension(OpenSSL::X509::Extension.new("crlNumber", crlnum))
      end
      alias serial= crl_number=
    end

    class Revoked
      UNSPECIFIED            = 0
      KEY_COMPROMISE         = 1
      CA_COMPROMISE          = 2
      AFFILIATION_CHANGED    = 3
      SUPERSEDED             = 4
      CESSATION_OF_OPERATION = 5
      CERTIFICATE_HOLD       = 6
      REMOVE_FROM_CRL        = 8
      PRIVILEGE_WITHDRAWN    = 9
      AA_COMPROMISE          = 10

      ReasonString = {
        UNSPECIFIED            => "Unspecified",
        KEY_COMPROMISE         => "Key Compromise",
        CA_COMPROMISE          => "CA Compromise",
        AFFILIATION_CHANGED    => "Affiliation Changed",
        SUPERSEDED             => "Superseded",
        CESSATION_OF_OPERATION => "Cessation Of Operation",
        CERTIFICATE_HOLD       => "Certificate Hold",
        REMOVE_FROM_CRL        => "Remove From CRL",
        PRIVILEGE_WITHDRAWN    => "Privilege Withdrawn",
        AA_COMPROMISE          => "AA Compromise",
      }

      Reasons = ReasonString.keys.sort

      def Revoked.generate(serial, time, reason=UNSPECIFIED)
        revoked = OpenSSL::X509::Revoked.new
        revoked.serial = serial
        revoked.time = time
        enum = OpenSSL::ASN1::Enumerated(reason)
        ext = OpenSSL::X509::Extension.new("CRLReason", enum)
        revoked.add_extension(ext)
        return revoked
      end

      def reason
        extensions.each{|ext|
          if ext.oid == "CRLReason"
            return Integer(ext.value)
          end
        }
      end

      def reason_string
        if n = reason
          return ReasonString[n]
        end
        return nil
      end
    end

    class Name
      def ==(other)
        return eql?(other)
      end
    end
  end

  class Config
    def extensions=(ext)
      ext.each{|oid, value|
        add_value("extensions", oid, value || "")
      }
    end

    def extensions
      ret = []
      each{|section, oid, value|
        ret << [oid, value] if section == "extensions"
      }
      return ret
    end

    def dup
      return OpenSSL::Config.parse(to_s)
    end

    def dump
      to_s
    end

    def method_missing(method, *args)
      method = method.to_s
      if /(\w+)=$/ =~ method
        if args.size == 1
          return add_value("default", $1, args[0].to_s)
        else
          return add_value("default", $1, args.to_s)
        end
      elsif args.empty?
        return get_value("default", method)
      else
        raise NoMethodError, "undefined method `#{method}' for #{inspect}"
      end
    end
  end

  class BN
    def hash
      to_i.hash
    end
  end
end
